#include "Database.h"
#include "Sequence.h"
#include "Level.h"

#include <Base\Containers\range_for_std_containers.hpp>
#include <Base\IO\FileUtil.hpp>
#include <Base\IO\MemoryReadStream.h>
#include <Base\IO\ReadStreamBinLex.h>
#include <Base\IO\ReadStreamBinLex_stdvector.h>
#include <Base\IO\ReadStreamZLib.h>
#include <Base\String\Formatting.h>
#include <Base\String\Trimming.h>

#include <unordered_map>

namespace hyplex::drakan
{
	String escapeToFilename(string_view s)
	{
		String out;
		out.reserve(s.size());
		// Windows doesn't like trailing spaces in file names.
		const size_t trailingWhitespaceI = rtrimmed(s).size();
		for (size_t i = 0; i < s.size(); ++i)
		{
			const char c = s[i];

			if (c >= 0 && (isalnum(c) || (c == ' ' && i < trailingWhitespaceI)))
				out += c;
			else
			{
				switch (c)
				{
				case '(':
				case ')':
				case '[':
				case ']':
				case ',':
				//case '.': // Trailing dots are not allowed.
				case '-':
				case '_':
				case '=':
				case '+':
					out += c;
					break;
				default:
					out += '{';
					printfmt(out, "%02x", (unsigned)(uint8_t)c);
					out += '}';
					break;
				}
			}
		}
		return out;
	}

	void INode::dumpNode(const NodeDumpArg&) const
	{
	}

	ENodeType Directory::getType() const noexcept
	{
		return ENodeType::DIRECTORY;
	}
	void Directory::dumpNode(const NodeDumpArg& arg) const
	{
		const String childrenDir = name.empty() ? arg.dir : arg.dir + escapeToFilename(name) + '\\';
		for (const auto& child : children)
			child->dumpNode({ arg.dbMeta, arg.db, arg.outputRootDir, arg.dbRootDir, childrenDir });
		Texture::dumpAnimationInfo(*this, childrenDir);
	}

	String Database::getEntriesDump() const
	{
		String s;
		for (const Entry& entry : entries)
		{
			printfmt(s, "%02x %02x %04x %04x @%u %i bytes: ", entry.type, entry.subtype, entry.unk00, entry.parentDirectory, entry.offset, entry.size);

			if (entry.type == 0x40 && entry.subtype == 0x00)
			{
				const array_view<const char> header = asArrayView(blob).slice_n(entry.offset, 50);
				for (unsigned int i = 0; i < header.size(); ++i)
				{
					if (i && i % 4 == 0)
						s += ' ';
					printfmt(s, "%02x ", (uint8_t)header[i]);
				}
			}

			s += '\n';
		}
		
		return s;
	}

	std::optional<Database> loadDatabase(std::string_view inputPath)
	{
		const std::optional<String> fileData = loadEntireFileAsStringIfExists(string_view(inputPath.data(), inputPath.size()));

		if (!fileData)
			return std::nullopt;

		BasicBufferedLexState bin{ *fileData };

		const auto header = read<Header>(bin);

		BasicBufferedLexState directoryBin{ asArrayView(*fileData).slice(header.directoryOffset) };

		Database db;

		db.blob.assign(fileData->begin(), fileData->end());
		readArray(directoryBin, db.entries, header.numEntries);

		return db;
	}

	String trimNul(String str)
	{
		return rtrimmed(std::move(str), [](char c) { return c == '\0'; });
	}

	void Database::parse()
	{
		std::vector<std::pair<std::unique_ptr<INode>, unsigned int>> nodeParentDirs;
		
		std::unordered_map<unsigned int, Directory*> dirsById;

		std::unordered_map<unsigned int, Model*> modelsById;
		std::unordered_map<unsigned int, Animation*> animsById;

		const auto addToParent = [&](std::unique_ptr<INode> node, unsigned int parentId) {
			if (parentId)
				nodeParentDirs.emplace_back(std::move(node), parentId);
			else
				rootDir.children.push_back(std::move(node));
		};

		for (const Entry& entry : entries)
		{
			// 00 00 - Level Metadata
			// 01 00 - Level Terrain
			// 02 00 - Level Layer Groups
			// 03 00 - 20000?
			// 20 00 - Level Entities

			// 30 00 uint16_t n, uint32_t[n] RGBX Palette
			// 40 00 Texture

			// 00 01 dragon.rfl path
			// 01 01 Object directory
			// 02 01 Object

			// 00 02 model info
			// 03 02 vertex data
			// 04 02 polygon data
			// 06 02 texture refs
			// 07 02 probably something to do with texture animation. UPDATE: Nope. It's bounding spheres. See FX's IceShot(H).
			// 08 02 lod info/scene nodes/animation refs
			// 09 02 uint16, uint32. Haven't found a non-zero occurence yet.

			// 02 03 Sound
			// 12 03 Sequence directory
			// 11 03 Sequence
			// 10 03 uint32 unk00, unk04
			
			// 02 04 uint32

			// 01 05 Animation info
			// 02 05 Animation frame index ranges
			// 03 05 Animation frames

			

			BasicBufferedLexState bin{ asArrayView(blob).slice_n(entry.offset, entry.size) };

			if ((entry.type == 1 && entry.subtype != 5 && entry.subtype != 0) || (entry.type == 0x12 && entry.subtype == 3))
			{
				auto newDir = std::make_unique<Directory>();
				newDir->name = trimNul(readString(bin, read<uint16_t>(bin)));
				
				dirsById.emplace(entry.unk00, newDir.get());
				addToParent(std::move(newDir), entry.parentDirectory);
			}
			else if (entry.subtype == 0 && entry.type == 0x30)
			{ // Texture palette.
				readArray(bin, palette, read<uint16_t>(bin));
			}
			else if (entry.subtype == 0 && entry.type == 0x40)
			{
				auto texture = std::make_unique<Texture>();
				texture->consumeData(entry, bin);
				addToParent(std::move(texture), entry.parentDirectory);
			}
			else if (entry.subtype == 2 && entry.unk00 != 0) // Model databases have a null model at the end for some reason.
			{
				Model*& model = modelsById[entry.unk00];
				if (!model)
				{
					auto newNode = std::make_unique<Model>();
					model = newNode.get();
					model->id = entry.unk00;
					addToParent(std::move(newNode), entry.parentDirectory);
				}

				model->consumeData(entry, bin);
			}
			else if (entry.subtype == 3)
			{
				if (entry.type == 2)
				{
					auto newNode = std::make_unique<Sound>();
					newNode->id = entry.unk00;
					newNode->name = trimNul(readString(bin, read<uint16_t>(bin)));
					newNode->unk = read<uint32_t>(bin);
					newNode->numChannels = read<uint16_t>(bin);
					newNode->bitsPerSample = read<uint16_t>(bin);
					newNode->sampleRate = read<uint32_t>(bin);
					newNode->volume = read<int32_t>(bin);
					newNode->dropOffDistance = read<float>(bin);
					newNode->priority = read<uint32_t>(bin);
					const uint32_t n = read<uint32_t>(bin);
					if (!read<uint32_t>(bin))
						readArray(bin, newNode->blob, n);
					else // Compressed.
					{
						const uint32_t compressedSize = read<uint32_t>(bin);
						std::vector<char> zlibData;
						readArray(bin, zlibData, compressedSize);

						ZLibBytesStreamFromChunksStream<MemoryReadStream> zlibBin(MemoryReadStream(asArrayView(zlibData)));

						newNode->blob.resize(n);
						zlibBin.read(newNode->blob);
					}
					
					addToParent(std::move(newNode), entry.parentDirectory);
				}
				else if (entry.type == 0x11)
				{
					auto obj = std::make_unique<Sequence>();
					obj->id = entry.unk00;
					obj->consumeData(entry, bin);
					addToParent(std::move(obj), entry.parentDirectory);
				}
			}
			else if (entry.subtype == 5 && entry.type != 0) // type 0 seems to be database meta info
			{
				Animation*& anim = animsById[entry.unk00];
				if (!anim)
				{
					auto newNode = std::make_unique<Animation>();
					anim = newNode.get();
					anim->id = entry.unk00;
					addToParent(std::move(newNode), entry.parentDirectory);
				}

				anim->consumeData(entry, bin);
			}
			else if (entry.subtype == 1 && entry.type == 2)
			{
				auto obj = std::make_unique<Object>();
				obj->id = entry.unk00;
				obj->consumeData(entry, bin);
				addToParent(std::move(obj), entry.parentDirectory);
			}
			else if (entry.subtype == 0)
			{
				if (entry.type == 0)
				{
					auto obj = std::make_unique<LevelMetadata>();
					obj->consumeData(entry, bin);
					addToParent(std::move(obj), entry.parentDirectory);
				}
				else if (entry.type == 1)
				{
					auto obj = std::make_unique<LevelTerrain>();
					obj->consumeData(entry, bin);
					addToParent(std::move(obj), entry.parentDirectory);
				}
				else if (entry.type == 0x20)
				{
					auto obj = std::make_unique<LevelEntityCollection>();
					obj->consumeData(entry, bin);
					addToParent(std::move(obj), entry.parentDirectory);
				}
			}
		}

		for (auto& kv : nodeParentDirs)
		{
			if (const auto it = dirsById.find(kv.second); it != dirsById.end())
				it->second->children.push_back(std::move(kv.first));
			else
				rootDir.children.push_back(std::move(kv.first));
		}
	}

	
	

}