#include "Image.h"

#include <Base\Exception.h>
#include <Base\CallInDtor.hpp>

#define PNG_USE_DLL 1
#include <png.h>

namespace hyplex::gfx
{
	static void PNGAPI writeErrorHandler(png_struct*, const char* msg)
	{
		// This should be okay. If libpng can longjmp from somewhere, an exception shouldn't break anything.
		throw RuntimeError("PNG write error: "_sv + cstringView(msg));
	}

	std::vector<char> makePNG(const IImage& img)
	{
		// http://www.libpng.org/pub/png/book/chapter15.html
		HYPLEX_ENFORCE(img.getPixelFormat() == EPixelFormat::U8RGBA);

		png_struct* png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, nullptr, writeErrorHandler, nullptr);
		if (!png_ptr)
			throw RuntimeError("PNG write error. png_create_write_struct failed.");

		png_info* info_ptr = png_create_info_struct(png_ptr);

		const auto pngDTor = make_call_in_dtor([&] { png_destroy_write_struct(&png_ptr, &info_ptr); });

		if (!info_ptr)
			throw RuntimeError("PNG write error. png_create_info_struct failed.");

		std::vector<char> data;

		static const auto write = [](png_struct* png_ptr, uint8_t* src, png_size_t size) {
			auto* const data = static_cast<std::vector<char>*>(png_get_io_ptr(png_ptr));
			data->insert(data->end(), reinterpret_cast<const char*>(src), reinterpret_cast<const char*>(src) + size);
		};

		static const auto flush = [](png_struct*) {
		};

		png_set_write_fn(png_ptr, &data, write, flush);
		png_set_compression_level(png_ptr, 9);

		png_set_IHDR(png_ptr, info_ptr, img.getDimensions().x, img.getDimensions().y,
			8, PNG_COLOR_TYPE_RGB_ALPHA, PNG_INTERLACE_NONE,
			PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

		png_write_info(png_ptr, info_ptr);

		const unsigned int pitch = img.getBytesPerRow();
		for (unsigned int y = 0, h = img.getDimensions().y; y < h; ++y)
			png_write_row(png_ptr, img.getRawData().slice_n(y * pitch, pitch).data());

		png_write_end(png_ptr, nullptr);

		return data;
	}
}