#include "Level.h"

#include <Base\Containers\range_for_std_containers.hpp>
#include <Base\IO\FileUtil.hpp>
#include <Base\IO\MemoryReadStream.h>
#include <Base\IO\ReadStreamBinLex.h>
#include <Base\IO\ReadStreamBinLex_stdvector.h>
#include <Base\IO\ReadStreamZLib.h>
#include <Base\Math\math_strfmt.h>
#include <Base\String\Formatting.h>

#include <iostream>

namespace hyplex::drakan
{
	static const bool kEnableDump = true;

	void LevelMetadata::consumeData(const Entry& entry, BasicBufferedLexState& bin)
	{
		if (entry.type != 0 || entry.subtype != 0)
			return;

		name = trimNul(readString(bin, read<uint16_t>(bin)));
		maxWidth = read<uint32_t>(bin);
		maxHeight = read<uint32_t>(bin);

		const size_t numDBs = read<uint32_t>(bin);
		dbRefs.reserve(numDBs);
		for (size_t i = 0; i < numDBs; ++i)
		{
			const unsigned int id = read<uint32_t>(bin);
			dbRefs.push_back({ id, trimNul(readString(bin, read<uint16_t>(bin))) });
		}
	}
	ENodeType LevelMetadata::getType() const noexcept
	{
		return ENodeType::LEVEL_METADATA;
	}
	void LevelMetadata::dumpNode(const NodeDumpArg& arg) const
	{
		if (!kEnableDump)
			return;

		String str;

		printfmt(str, "name %s\n", name);
		printfmt(str, "maxWidth %u\n", maxWidth);
		printfmt(str, "maxHeight %u\n", maxHeight);
		for(const DatabaseRef& dbRef: dbRefs)
			printfmt(str, "DBRef %u %s\n", dbRef.id, dbRef.path);

		FileUnbufferedWriteStream(sprintfmt("%sLevelMetadata %s.txt", arg.dir, escapeToFilename(name)), EFileWriteOpenMode::RECREATE, true)
			.write(asRawUChars(asArrayView(str)));
	}

	void LevelTerrain::consumeData(const Entry& entry, BasicBufferedLexState& bin)
	{
		if (entry.type != 1 || entry.subtype != 0)
			return;
		
		const unsigned int numLayers = read<uint32_t>(bin);
		layers.reserve(numLayers);

		for (size_t i = 0; i < numLayers; ++i)
		{
			Layer layer{};
			layer.id = read<uint32_t>(bin);
			layer.size = read<uivec2>(bin);
			layer.type = EType(read<uint32_t>(bin));
			layer.position = read<uivec2>(bin);
			layer.worldHeight = read<float>(bin);
			layer.name =  trimNul(readString(bin, read<uint16_t>(bin)));
			const uint32_t unk0 = read<uint32_t>(bin);
			HYPLEX_ENFORCE(unk0 == 0 || unk0 == 0x2);
			layer.memberOfAlternateBlendingGroup = (unk0 & 0x2) != 0;
			layer.lightDirection = read<float>(bin);
			layer.lightAscension = read<float>(bin);
			layer.lightDiffuseColour = read<std::array<uint8_t, 3>>(bin);
			skipUptoN(bin, 1);
			layer.lightAmbientColour = read<std::array<uint8_t, 3>>(bin);
			skipUptoN(bin, 1);
			std::swap(layer.lightDiffuseColour[0], layer.lightDiffuseColour[2]);
			std::swap(layer.lightAmbientColour[0], layer.lightAmbientColour[2]);
			//const uint32_t unk1 = read<uint32_t>(bin);
			//if (unk1 != 0)
			//	std::cout << layer.name << " " << std::hex << unk1 << std::endl;
			layer.unk0 = read<uint32_t>(bin);
			readArray(bin, layer.visibleLayerIndices, read<uint32_t>(bin));

			layers.push_back(std::move(layer));
		}

		// ??
		HYPLEX_ENFORCE(read<uint32_t>(bin) == 1);

		using RawVertex = LayerVertex;

#pragma pack(push)
#pragma pack(1)
		struct RawCell
		{
			uint16_t flags;
			uint32_t textureA;
			uint32_t textureB;
			ui16vec2 texCoords[4];
		};
#pragma pack(pop)

		std::vector<char> decompressedData;
		std::vector<RawCell> rawCells;

		for (size_t i = 0; i < numLayers; ++i)
		{
			

			Layer& layer = layers[i];
			const size_t numVertices = (layer.size.x + 1) * (layer.size.y + 1);
			const size_t numCells = layer.size.x * layer.size.y;
			const size_t decompressedSize = numVertices * sizeof(RawVertex) + numCells * sizeof(RawCell);

			if (decompressedData.size() < decompressedSize)
				decompressedData.insert(decompressedData.end(), decompressedSize - decompressedData.size(), 0);

			const array_view<const char> compressedData = consumeUptoN(bin, read<uint32_t>(bin));

			ZLibBytesStreamFromChunksStream<MemoryReadStream> zlibStream{ MemoryReadStream(compressedData) };
			HYPLEX_ENFORCE(zlibStream.read(decompressedData) == decompressedSize);

			BasicBufferedLexState terrainBin{ asArrayView(decompressedData).slice_n(0, decompressedSize) };
			readArray(terrainBin, layer.vertices, numVertices);
			readArray(terrainBin, rawCells, numCells);

			layer.cells.reserve(numCells);

			for (const RawCell& rawCell: rawCells)
			{
				LayerGridCell cell{};
				cell.flags = rawCell.flags;
				for (unsigned int j = 0; j < 4; ++j)
				{
					const ui16vec2 texCoord = rawCell.texCoords[j];
					// This fails sometimes. Probably not important.
					//HYPLEX_ENFORCE((texCoord.x == 0 || texCoord.x == UINT16_MAX) && (texCoord.y == 0 || texCoord.y == UINT16_MAX));
					cell.texCoordXBits |= texCoord.x != 0 ? 1 << j : 0;
					cell.texCoordYBits |= texCoord.y != 0 ? 1 << j : 0;
				}
				cell.textureIds[0] = rawCell.textureA;
				cell.textureIds[1] = rawCell.textureB;

				layer.cells.push_back(cell);
			}
		}

		HYPLEX_ENFORCE(!peek(bin));
	}
	ENodeType LevelTerrain::getType() const noexcept
	{
		return ENodeType::LEVEL_TERRAIN;
	}
	void LevelTerrain::dumpNode(const NodeDumpArg& arg) const
	{
		if (!kEnableDump)
			return;

		String str;

		for (size_t i = 0; i < layers.size(); ++i)
		{
			str.clear();

			const Layer& layer = layers[i];

			const auto dumpVertexGrid = [&](auto f) {
				for (unsigned int y = 0; y < layer.size.y + 1; ++y)
				{
					str += '\t';
					for (unsigned int x = 0; x < layer.size.x + 1; ++x)
						f(layer.vertices[y * (layer.size.x + 1) + x]);
					str += '\n';
				}
			};

			const auto dumpCellGrid = [&](auto f) {
				for (unsigned int y = 0; y < layer.size.y; ++y)
				{
					str += '\t';
					for (unsigned int x = 0; x < layer.size.x; ++x)
						f(layer.cells[y * layer.size.x + x]);
					str += '\n';
				}
			};

			printfmt(str, "index %u\n", i);
			printfmt(str, "id %u\n", layer.id);
			printfmt(str, "name %u\n", layer.name);
			printfmt(str, "unk0 %08x\n", layer.unk0);
			printfmt(str, "position %i\n", layer.position);
			printfmt(str, "size %i\n", layer.size);
			printfmt(str, "type %i\n", layer.type == FLOOR ? "floor"_sv : layer.type == CEILING ? "ceiling"_sv : "between"_sv);
			printfmt(str, "worldHeight %f\n", layer.worldHeight);
			printfmt(str, "lightDirection %f\n", layer.lightDirection);
			printfmt(str, "lightAscension %f\n", layer.lightAscension);
			printfmt(str, "lightDiffuseColour %f\n", asArrayView(layer.lightDiffuseColour));
			printfmt(str, "lightAmbientColour %f\n", asArrayView(layer.lightAmbientColour));
			printfmt(str, "visible layers %u\n", asArrayView(layer.visibleLayerIndices));
			printfmt(str, "vertex flags\n");
			dumpVertexGrid([&](const LayerVertex& v) {
				printfmt(str, "%x ", v.flags);
			});
			printfmt(str, "vertex heights\n");
			dumpVertexGrid([&](const LayerVertex& v) {
				printfmt(str, "% .4f ", (v.quantisedHeight + 32768) / 1024.0f);
			});
			printfmt(str, "cell divisions\n");
			dumpCellGrid([&](const LayerGridCell& c) {
				if (c.flags & LayerGridCell::kFlagIsFlat)
					str += "= "_sv;
				else if(c.flags & LayerGridCell::kFlagDivisionIsDownRight)
					str += "\\ "_sv;
				else
					str += "/ "_sv;
			});
			printfmt(str, "cell unk4\n");
			dumpCellGrid([&](const LayerGridCell& c) {
				if (c.flags & LayerGridCell::kFlagWalkable0)
					str += "+ "_sv;
				else
					str += "- "_sv;
			});
			printfmt(str, "cell unk8\n");
			dumpCellGrid([&](const LayerGridCell& c) {
				if (c.flags & LayerGridCell::kFlagWalkable1)
					str += "+ "_sv;
				else
					str += "- "_sv;
			});

			printfmt(str, "cell texcoords\n");
			String rows[4];
			for (unsigned int y = 0; y < layer.size.y; ++y)
			{
				rows[0].clear();
				rows[1].clear();
				rows[2].clear();
				rows[3].clear();

				for (unsigned int x = 0; x < layer.size.x; ++x)
				{
					const LayerGridCell& cell = layer.cells[y * layer.size.x + x];
					rows[0] += "+-----"_sv;
					rows[2] += "|     "_sv;
					printfmt(rows[1], "|%c%c %c%c",
						cell.texCoordXBits & 1 ? '1' : '0',
						cell.texCoordYBits & 1 ? '1' : '0',
						cell.texCoordXBits & 2 ? '1' : '0',
						cell.texCoordYBits & 2 ? '1' : '0');
					printfmt(rows[3], "|%c%c %c%c",
						cell.texCoordXBits & 4 ? '1' : '0',
						cell.texCoordYBits & 4 ? '1' : '0',
						cell.texCoordXBits & 8 ? '1' : '0',
						cell.texCoordYBits & 8 ? '1' : '0');
				}

				rows[0] += '+';
				rows[1] += '|';
				rows[2] += '|';
				rows[3] += '|';

				for (const string_view row : rows)
				{
					str += '\t';
					str += row;
					str += '\n';
				}
			}

			str += '\t';
			str += rows[0];
			str += '\n';

			

			FileUnbufferedWriteStream(sprintfmt("%sLevelLayers\\LevelLayer%u %s.txt", arg.dir, i, escapeToFilename(layer.name)), EFileWriteOpenMode::RECREATE, true)
				.write(asRawUChars(asArrayView(str)));
		}

		

		FileUnbufferedWriteStream(sprintfmt("%sLevelTerrain.txt", arg.dir), EFileWriteOpenMode::RECREATE, true)
			.write(asRawUChars(asArrayView(str)));
	}

	void LevelEntityCollection::consumeData(const Entry& entry, BasicBufferedLexState& bin)
	{
		if (entry.type != 0x20 || entry.subtype != 0)
			return;

		const unsigned int n = read<uint16_t>(bin);
		entities.reserve(n);
		for (unsigned int i = 0; i < n; ++i)
		{
#pragma pack(push)
#pragma pack(1)
			const struct EntityHeader
			{
				uint32_t id;
				uint32_t objectId;
				uint32_t lightingLayerId;
				fvec3 position;
				uint32_t flags;
				uint16_t unk1C;
				uint16_t unk1E;
			} header = read<EntityHeader>(bin);
#pragma pack(pop)

			Entity entity{};
			entity.id = header.id;
			entity.objectId = header.objectId;
			entity.position = header.position / 2048.0f;
			entity.lightingLayerId = header.lightingLayerId;
			HYPLEX_ENFORCE((header.flags & ~uint32_t(0x0101)) == 0);
			entity.visible = (header.flags & 0x0001) != 0;
			//HYPLEX_ENFORCE(header.unk1C == 1);
			entity.unk1C = header.unk1C;
			readArray(bin, entity.links, header.unk1E);
			entity.rotationDegrees = read<i16vec3>(bin);
			if (header.flags & 0x0100)
				read(bin, entity.scale);
			else
				entity.scale = { 1, 1, 1 };

			entity.props = Object::readProperties(bin, true);

			entities.push_back(std::move(entity));
		}

		HYPLEX_ENFORCE(!peek(bin));
	}
	ENodeType LevelEntityCollection::getType() const noexcept
	{
		return ENodeType::LEVEL_ENTITY_COLLECTION;
	}
	void LevelEntityCollection::dumpNode(const NodeDumpArg& arg) const
	{
		if (!kEnableDump)
			return;
		
		String str;

		for (size_t i = 0; i < entities.size(); ++i)
		{
			const Entity& entity = entities[i];
			printfmt(str, "Entity %u:\n", i);
			printfmt(str, "\tid %04x\n", entity.id);
			printfmt(str, "\tobject %04x\n", entity.objectId);
			if (entity.lightingLayerId)
				printfmt(str, "\tobject %04x\n", entity.lightingLayerId);
			if (!entity.visible)
				printfmt(str, "\thidden\n");
			if (entity.unk1C != 1)
				printfmt(str, "\tunk1C %u\n", entity.unk1C);
			printfmt(str, "\tposition %g\n", entity.position);
			printfmt(str, "\trotation %i\n", entity.rotationDegrees);
			if(entity.scale != fvec3(1, 1, 1))
				printfmt(str, "\tscale %f\n", entity.scale);
			if(!entity.links.empty())
				printfmt(str, "\tlinks %u\n", asArrayView(entity.links));

			for (const Object::Property& prop : entity.props)
				printfmt(str, "\tprop %u: %s\n", prop.index, prop.toString());
		}

		FileUnbufferedWriteStream(sprintfmt("%sLevelEntities.txt", arg.dir), EFileWriteOpenMode::RECREATE, true)
			.write(asRawUChars(asArrayView(str)));
	}
}