#include "RFL.h"

#include <Base\Containers\range_for_std_containers.hpp>
#include <Base\IO\FileUtil.hpp>
#include <Base\IO\ReadStreamBinLex.h>
#include <Base\IO\ReadStreamBinLex_stdvector.h>
#include <Base\String\Formatting.h>

namespace hyplex::drakan
{
	// Dragon.rfl
	// 
	// ROC OBJECTS
	// 
	// Level Editor read at 004712FE.
	// 
	// ROC OBJECT array size memory read at 00444C80, 004422DA
	// After class names, array of uint32_t read at 004420FE. "Section 2"
	// 
	// Format:
	// uint32_t n;
	// struct Unk
	// {
	// 	uint32_t unkId;
	// 	uint32_t category;
	// 	uint16_t unk08; // Read at 00444CD2
	// 	uint32_t unk0C;
	// 	uint32_t unk10;
	// 	uint32_t unk14; // Read at 00444CCF, 004450B0
	// 	uint32_t unk18; // Number of entries in "Section 2A".
	// 	uint32_t unk1C; // Offset to "Section 2A" from the start of this Unk. Entries in "Section 2A" are also offsets from this Unk that point into the same section.
	// } unk[n];
	// 
	// struct Section2AEntry
	// {
	// 	uint32_t offsetFromUnk; // Offset to Section2BEntry.
	// };
	// 
	// struct Section2BEntry
	// {
	// 	uint32_t offsetFromSection2BEntryToName;
	// 	uint32_t unk04; // Accumulated at 00442105. Probably a number of 16 byte entries. Added onto offsetFromUnk.
	// 	uint32_t offsetFromSection2BEntryToSection3Entry; // Read at 0044212A.
	// };
	// 
	// // Property
	// struct Section3Entry
	// {
	// 	uint32_t offsetFromSection3EntryToName;
	// 	uint32_t unkFlags04; // Read at 00442144. Flag 0x1000. Lower bits 0xFFF are a number. Seitch at 00442153 branches for 0x3..0xC.
	// 	union
	// 	{
	// 		uint32_t unk08;
	// 		float unk08; // Default value?
	// 	};
	// 	uint32_t unk0C;
	// };

	DataType::DataType(uint32_t rawTypeInfo) noexcept
	{
		elementType = EDataType(rawTypeInfo & 0xFFF);
		const uint8_t typeFlags = uint8_t(rawTypeInfo >> 12);
		isList = (typeFlags & 1) != 0;
		isObjectOfClassCategory = (typeFlags & 2) == 0;
	}

	ClassCategoryLookup loadRFLClassCategories(RawConstMemoryRange blob)
	{
		BasicBufferedLexState bin{ asRawChars(asArrayView(blob)) };
		const char* const blobStart = bin.buf.data();

		const unsigned int n = read<uint32_t>(bin);

		std::map<unsigned int, String> result;

		for (unsigned int i = 0; i < n; ++i)
		{
			const size_t addr = bin.buf.data() - blobStart;

			const auto relBlob = asRawChars(asArrayView(blob)).slice(addr);

			const unsigned int id = read<uint32_t>(bin);
			const unsigned int nameOffset = read<uint32_t>(bin);

			result.emplace(id, cstringView(relBlob.slice(nameOffset)));
		}

		return result;
	}

	std::map<unsigned int, Class> loadRFLClasses(RawConstMemoryRange blob)
	{
		std::map<unsigned int, Class> classes;

		BasicBufferedLexState bin{ asRawChars(asArrayView(blob)) };

		const char* const blobStart = bin.buf.data();

		const unsigned int numClasses = read<uint32_t>(bin);

		struct ClassInfo
		{
			uint32_t unkId;
			uint32_t category;
			uint16_t unk08; // Read at 00444CD2. Class version number?
			uint32_t unk0C;
			uint32_t offsetToName;
			uint32_t unk14; // Read at 00444CCF, 004450B0. Number of props.
			uint32_t numPropSections; // Number of entries in "Section 2A".
			uint32_t offsetToPropSections; // Offset to "Section 2A" from the start of this Unk. Entries in "Section 2A" are also offsets from this Unk that point into the same section.
		};

		struct PropertySectionInfo
		{
			uint32_t offsetFromPropertySectionInfoToName;
			uint32_t numProps;
			uint32_t offsetFromPropertySectionInfoToProps;
		};

		struct Property
		{
			uint32_t offsetFromPropertyToName;
			uint32_t type;
			union
			{
				int32_t i32;
				float f32;
				std::array<uint8_t, 4> argb;
			} defaultValue;
			uint32_t unk0C; // class type?
		};

		String out;

		for (unsigned int i = 0; i < numClasses; ++i)
		{
			const size_t classAddr = bin.buf.data() - blobStart;

			const ClassInfo classInfo = read<ClassInfo>(bin);

			const auto classRelBlob = asRawChars(asArrayView(blob)).slice(classAddr);

			Class klass;

			klass.id = classInfo.unkId;
			klass.name = cstringView(classRelBlob.slice(classInfo.offsetToName));
			klass.categoryId = classInfo.category;
			klass.unk08 = classInfo.unk08;
			klass.unk0C = classInfo.unk0C;
			klass.unk14 = classInfo.unk14;

			BasicBufferedLexState propSectionOffsetsBin{ classRelBlob.slice(classInfo.offsetToPropSections) };

			for (unsigned int propSectionI = 0; propSectionI < classInfo.numPropSections; ++propSectionI)
			{
				const size_t propSectionAddr = classAddr + read<uint32_t>(propSectionOffsetsBin);
				const auto propSectionRelBlob = asRawChars(asArrayView(blob)).slice(propSectionAddr);
				BasicBufferedLexState propSectionBin{ propSectionRelBlob };

				const PropertySectionInfo rawPropSection = read<PropertySectionInfo>(propSectionBin);

				Class::PropertySection propSection;
				propSection.name = cstringView(propSectionRelBlob.slice(rawPropSection.offsetFromPropertySectionInfoToName));
				propSection.propsIR = index_range_n(klass.props.size(), 0);

				const auto propsBlob = propSectionRelBlob.slice(rawPropSection.offsetFromPropertySectionInfoToProps);

				BasicBufferedLexState propsBin{ propsBlob };

				for (unsigned int propI = 0; propI < rawPropSection.numProps; ++propI)
				{
					const auto propRelBlob = propsBin.buf;

					const Property rawProp = read<Property>(propsBin);

					Class::Property prop;
					prop.name = cstringView(propRelBlob.slice(rawProp.offsetFromPropertyToName));
					prop.objClassCategoryId = rawProp.unk0C;
					//prop.unkTypeHighBits = uint8_t(rawProp.type >> 12);
					prop.type = DataType(rawProp.type);

					//if (prop.name == "Player Slot")
					//	_CrtDbgBreak();
	
					switch (prop.type.elementType)
					{
					case EDataType::INT:
					case EDataType::CHOICE:
						prop.defaultValue.i32 = rawProp.defaultValue.i32;
						break;
					case EDataType::FLOAT:
						prop.defaultValue.f32 = rawProp.defaultValue.f32;
						break;
					case EDataType::COLOUR:
						prop.defaultValue.argb = rawProp.defaultValue.argb;
						break;
					}

					klass.props.push_back(std::move(prop));
				}

				propSection.propsIR = index_range(propSection.propsIR.begin(), klass.props.size());
				klass.propertySections.push_back(std::move(propSection));
			}

			classes.emplace(klass.id, std::move(klass));
		}

		//FileUnbufferedWriteStream("RFL Objects Dump.txt", EFileWriteOpenMode::RECREATE)
		//	.write(asRawUChars(asArrayView(out)));

		return classes;
	}

	String Class::toString(const ClassCategoryLookup& classCats, const ClassLookup& classes) const
	{
		String out;
		printfmt(out, "Class %02x %s (%08x %08x %08x) : ", id, name, unk08, unk0C, unk14);

		if (const auto it = classCats.find(categoryId); it != classCats.end())
			out += it->second;
		else
			out += "[invalid class category id]";
		out += '\n';
		
		for (const PropertySection& section : propertySections)
		{
			printfmt(out, "\t%s\n", section.name);
			for (const Property& prop : section.propsIR.in(props))
			{
				printfmt(out, "\t\t%s = ", prop.name);

				switch (prop.type.elementType)
				{
				case EDataType::INT:
					printfmt(out, "%i", prop.defaultValue.i32);
					break;
				case EDataType::FLOAT:
					printfmt(out, "%f", prop.defaultValue.f32);
					break;
				case EDataType::OBJECT:
					if (prop.type.isObjectOfClassCategory)
					{
						if (const auto it = classCats.find(prop.objClassCategoryId); it != classCats.end())
							printfmt(out, "[object of class category %s]", it->second);
						else if(prop.objClassCategoryId)
							printfmt(out, "[object of invalid class category %u]", prop.objClassCategoryId);
						else
							out += "[object of any class category]";
					}
					else
					{
						if (const auto it = classes.find(prop.objClassId); it != classes.end())
							printfmt(out, "[object of class %s]", it->second.name);
						else if(prop.objClassId)
							printfmt(out, "[object of invalid class %02x]", prop.objClassId);
						else
							out += "[object of any class]";
					}
					break;
				case EDataType::MODEL:
					printfmt(out, "[model]");
					break;
				case EDataType::SOUND:
					printfmt(out, "[sound]");
					break;
				case EDataType::CHOICE:
					printfmt(out, "%b", !!prop.defaultValue.i32);
					break;
				case EDataType::ANIM_CHANNEL:
					printfmt(out, "[animation channel]");
					break;
				case EDataType::ANIMATION:
					printfmt(out, "[animation]");
					break;
				case EDataType::STRING:
					printfmt(out, "[string]");
					break;
				case EDataType::SEQUENCE:
					printfmt(out, "[sequence]");
					break;
				case EDataType::LANDSCAPE_LAYER:
					printfmt(out, "[landscape layer]");
					break;
				case EDataType::TEXTURE:
					printfmt(out, "[texture]");
					break;
				case EDataType::COLOUR:
					printfmt(out, " RGBA %u %u %u %u", prop.defaultValue.argb[1], prop.defaultValue.argb[2], prop.defaultValue.argb[3], prop.defaultValue.argb[0]);
					break;
				default:
					HYPLEX_ENFORCE_UNREACHABLE();
				}

				if (prop.type.isList)
					out += " list"_sv;

				out += '\n';
			}
		}

		return out;

	}
}