#include "Sequence.h"

#include <Base\Containers\range_for_std_containers.hpp>
#include <Base\IO\FileUtil.hpp>
#include <Base\IO\MemoryReadStream.h>
#include <Base\IO\ReadStreamBinLex.h>
#include <Base\IO\ReadStreamBinLex_stdvector.h>
#include <Base\Math\math_strfmt.h>
#include <Base\String\Formatting.h>

namespace hyplex::drakan
{
	static const bool kEnableDump = true;

	void Sequence::consumeData(const Entry& entry, BasicBufferedLexState& bin)
	{
		if (entry.type == 0x11 && entry.subtype == 3 && entry.unk00 == 0x00c9)
		{
			name = trimNul(readString(bin, read<uint16_t>(bin)));

			struct Header
			{
				uint32_t unk00; // flags?
				uint32_t unk04;
				uint32_t unk08;
				uint32_t numActors;
			};

#pragma pack(push)
#pragma pack(1)
			struct RawActionBase
			{
				uint16_t type;
				float t;
			};
#pragma pack(pop)

			

			

			const Header header = read<Header>(bin);

			for (unsigned int i = 0; i < header.numActors; ++i)
			{
				Actor actor{};
				actor.name = trimNul(readString(bin, read<uint16_t>(bin)));

//#pragma pack(push)
//#pragma pack(1)
				struct RawActor
				{
					uint32_t id;
					uint32_t unk04;
					uint32_t unk08;
					uint32_t levelEntityId; // probably
					uint32_t numActions;
					//uint16_t unk14;
					//uint32_t unk16;
					//uint32_t unk1A;
				};
//#pragma pack(pop)

				const auto actorHeader = read<RawActor>(bin);

				actor.id = actorHeader.id;
				actor.levelEntityId = actorHeader.levelEntityId;

				HYPLEX_ENFORCE(actorHeader.unk04 == 1);
				HYPLEX_ENFORCE(actorHeader.unk08 == 0);
				//HYPLEX_ENFORCE(actorHeader.unk14 == 5);
				//HYPLEX_ENFORCE(actorHeader.unk16 == 0);
				//HYPLEX_ENFORCE(actorHeader.unk1A == 0);

				for (unsigned int j = 0; j < actorHeader.numActions; ++j)
				{
					// 0 = xform keyframe
					// 3 = sound

					const auto actionHeader = read<RawActionBase>(bin);

					Action action{};
					action.t = actionHeader.t;

					switch (actionHeader.type)
					{
					case 0:
					{
						const struct RawXFormKeyFrame
						{
							fvec4 rotationQuat;
							fvec3 position;
							uint32_t flags; // probably
							uint16_t interpolation; // probably
							uint16_t relativeTo; // probably
						} rawInfo = read<RawXFormKeyFrame>(bin);
						XFormKeyframe info{};
						info.rotation = fquat(rawInfo.rotationQuat);
						info.translation = rawInfo.position / 2048.0f;
						HYPLEX_ENFORCE(rawInfo.flags == 0);
						HYPLEX_ENFORCE(rawInfo.relativeTo == 0);
						switch (rawInfo.interpolation)
						{
						case 0x00: info.interpolation = EInterpolation::NONE; break;
						case 0x10: info.interpolation = EInterpolation::LINEAR_LINEAR; break;
						case 0x30: info.interpolation = EInterpolation::SINE_SPLINE; break;
						default:
							HYPLEX_ENFORCE_UNREACHABLE();
						}
						action.info = info;
						break;
					}
					case 1:
					{
						const struct RawAnimationAction
						{
							uint32_t channelI;
							uint32_t animId;
							uint32_t unk08;
							float transitionTime;
							float speed;
							uint32_t rootNodeTranslationModeFlags;
						} rawInfo = read<RawAnimationAction>(bin);

						HYPLEX_ENFORCE(rawInfo.unk08 == 0);
						HYPLEX_ENFORCE(rawInfo.rootNodeTranslationModeFlags <= 0b111111);

						AnimationAction info{};
						info.channelI = rawInfo.channelI;
						info.animId = rawInfo.animId;
						info.transitionTime = rawInfo.transitionTime;
						info.speed = rawInfo.speed;
						for (unsigned int k = 0; k < 3; ++k)
							info.rootNodeTranslationMode[k] = ERootNodeTranslationMode((rawInfo.rootNodeTranslationModeFlags >> (k * 2)) & 0b11);
						action.info = info;
						break;
					}
					case 3:
					{
						const struct RawSoundAction
						{
							uint32_t soundId;
							float duration;
							uint16_t volume;
							uint16_t flags;
						} rawInfo = read<RawSoundAction>(bin);
						SoundAction info{};
						info.duration = rawInfo.duration;
						info.soundId = rawInfo.soundId;
						HYPLEX_ENFORCE(rawInfo.volume <= 10000);
						info.volume = rawInfo.volume / 10000.0f;

						// 0x8 = flush from memory after playing
						HYPLEX_ENFORCE((rawInfo.flags & ~uint16_t(0x38)) == 0);
						info.mono = (rawInfo.flags & 0x10) != 0;
						info.notifyActor = (rawInfo.flags & 0x30) != 0;
						action.info = info;
						break;
					}

					case 4:
						action.info = read<AttachActorAction>(bin);
						break;

					case 5:
					{
						const uint32_t state = read<uint32_t>(bin);
						HYPLEX_ENFORCE(state == 0 || state == 1);
						action.info = RunStopAI{ state == 1 };
						break;
					}
					case 6:
					{
						const uint32_t state = read<uint32_t>(bin);
						HYPLEX_ENFORCE(state == 0 || state == 1);
						action.info = ShowHideAction{ state == 1 };
						break;
					}
					case 7:
					{
						const struct RawMessageAction
						{
							uint32_t message;
							uint32_t unk04;
						} rawInfo = read<RawMessageAction>(bin);
						MessageAction info{};
						info.message = rawInfo.message;
						HYPLEX_ENFORCE(rawInfo.unk04 == 0);
						action.info = info;
						break;
					}
					default:
						HYPLEX_ENFORCE_UNREACHABLE();
					}

					actor.actions.push_back(std::move(action));
				}

				actors.push_back(std::move(actor));
			}
		}
	}
	ENodeType Sequence::getType() const noexcept
	{
		return ENodeType::SEQUENCE;
	}
	void Sequence::dumpNode(const NodeDumpArg& arg) const
	{
		if (!kEnableDump)
			return;

		String str;

		printfmt(str, "id %u\n", id);
		printfmt(str, "name %s\n", name);
		str += '\n';

		struct ActionDumper
		{
			String& str;

			//std::variant<std::monostate, XFormKeyframe, SoundAction, RunStopAI, AnimationAction, AttachActorAction, MessageAction,
			//	ShowHideAction> info;
			void operator()(std::monostate)
			{
				str += "[none]";
			}

			void operator()(const XFormKeyframe& a)
			{
				printfmt(str, "XFormKeyframe %g %g", a.translation, a.rotation);
				if (a.ignorePosition)
					str += " ignorePosition"_sv;
				if (a.ignoreRotation)
					str += " ignoreRotation"_sv;

				switch (a.relative)
				{
				case ERelativeTo::THE_WORLD: str += " ERelativeTo::THE_WORLD"_sv; break;
				case ERelativeTo::ANOTHER_ACTOR: str += " ERelativeTo::ANOTHER_ACTOR"_sv; break;
				case ERelativeTo::LOOKAT_ACTOR: str += " ERelativeTo::LOOKAT_ACTOR"_sv; break;
				default:
					HYPLEX_ENFORCE_UNREACHABLE();
				}

				switch (a.interpolation)
				{
				case EInterpolation::NONE: str += " EInterpolation::NONE"_sv; break;
				case EInterpolation::LINEAR_LINEAR: str += " EInterpolation::LINEAR_LINEAR"_sv; break;
				case EInterpolation::LINEAR_SPLINE: str += " EInterpolation::LINEAR_SPLINE"_sv; break;
				case EInterpolation::SINE_SPLINE : str += " EInterpolation::SINE_SPLINE"_sv; break;
				default:
					HYPLEX_ENFORCE_UNREACHABLE();
				}
			}

			void operator()(const SoundAction& a)
			{
				printfmt(str, "SoundAction play %04x for %gs at volume %g", a.soundId, a.duration, a.volume);
				if (a.mono)
					str += " mono"_sv;
				if (a.notifyActor)
					str += " notifyActor";
			}

			void operator()(const RunStopAI& a)
			{
				printfmt(str, "RunStopAI %b", a.runAI);
			}

			void operator()(const AnimationAction& a)
			{
				printfmt(str, "AnimationAction play %04x on channel %u at speed %g", a.animId, a.channelI, a.speed);
				if (a.transitionTime)
					printfmt(str, ", transitionTime %g", a.transitionTime);
				if (a.channelI == 0)
				{
					str += ", root node translation mode XYZ"_sv;
					for (int i = 0; i < 3; ++i)
					{
						switch (a.rootNodeTranslationMode[i])
						{
						case ERootNodeTranslationMode::MESH: str += " MESH"_sv; break;
						case ERootNodeTranslationMode::FRAME: str += " FRAME"_sv; break;
						case ERootNodeTranslationMode::NONE: str += " NONE"_sv; break;
						case ERootNodeTranslationMode::INVALID: str += " INVALID"_sv; break;
						default:
							HYPLEX_ENFORCE_UNREACHABLE();
						}
					}
				}
			}

			void operator()(const AttachActorAction& a)
			{
				printfmt(str, "AttachActorAction attach actor %u to channel %u", a.attachedActorId, a.thisActorsChannelId);
			}

			void operator()(const MessageAction& a)
			{
				printfmt(str, "MessageAction %u", a.message);
			}

			void operator()(const ShowHideAction& a)
			{
				printfmt(str, "ShowHideAction %b", a.visible);
			}

		} actionDumper{ str };

		for (const Actor& actor : actors)
		{
			printfmt(str, "Actor %u %s - level entity %u\n", actor.id, actor.name, actor.levelEntityId);

			for (const Action& action : actor.actions)
			{
				printfmt(str, "\t%.7f: ", action.t);
				std::visit(actionDumper, action.info);
				str += '\n';
			}

			str += '\n';
		}
		
		FileUnbufferedWriteStream(sprintfmt("%sSequence%u %s.txt", arg.dir, id, name), EFileWriteOpenMode::RECREATE, true)
			.write(asRawUChars(asArrayView(str)));
	}
}